'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.extendModel = exports.handleWithContext = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/* utils start */
var cache = new Map();

var setCache = function setCache(path, value) {
    path.reduce(function (level, key, index) {
        if (path.length - index > 1) {
            !level.get(key) && level.set(key, new Map());
            return level.get(key);
        }
        level.set(key, value);
    }, cache);
};

var getCache = function getCache(path) {
    return path.reduce(function (level, key) {
        if (!level || !level.has(key)) {
            return;
        }
        return level.get(key);
    }, cache);
};
/* utils end */

var parseModelConfig = function parseModelConfig(model) {
    var ret = {};
    if (typeof model === 'string') {
        ret.namespace = model;
    } else {
        ret = model;
    }
    // 检查prefix
    (0, _invariant2.default)(!ret.prefix || /\$$/.test(ret.prefix), '[connectWithContext] model.prefix must end with `$`.');
    return ret;
};
var getModelConfig = function getModelConfig(modelContext) {
    return Object.keys(modelContext).reduce(function (ret, key) {
        // 检查namespace
        (0, _invariant2.default)(/^\$/.test(key), '[connectWithContext] Abstract model\'s namespace must start with `$`.');
        ret[key] = parseModelConfig(modelContext[key]);
        return ret;
    }, {});
};
var wrapDispatchWithModelContext = function wrapDispatchWithModelContext(dispatch, modelConfig) {
    return function (action) {
        var type = action.type;

        var _type$split = type.split('/'),
            _type$split2 = _slicedToArray(_type$split, 2),
            namespace = _type$split2[0],
            reducer = _type$split2[1];

        if (modelConfig[namespace]) {
            type = modelConfig[namespace].namespace + '/' + (modelConfig[namespace].prefix || '') + reducer;
        }
        dispatch(_extends({}, action, {
            type: type
        }));
    };
};
var proxyStateForModel = function proxyStateForModel(st, modelConfig) {
    var state = _extends({}, st);
    Object.keys(modelConfig).forEach(function (key) {
        var _modelConfig$key = modelConfig[key],
            namespace = _modelConfig$key.namespace,
            prefix = _modelConfig$key.prefix;

        !state.hasOwnProperty(key) && Object.defineProperty(state, key, {
            get: function get() {
                var modelState = state[namespace];
                // 有前缀的话
                if (prefix) {
                    modelState = Object.keys(modelState).reduce(function (ret, field) {
                        !field.indexOf(prefix) && (ret[field.substr(prefix.length)] = modelState[field]);
                        return ret;
                    }, {});
                }

                return modelState;
            }
        });
    });
    return state;
};
var wrapGetStateWithModelContext = function wrapGetStateWithModelContext(getState, modelConfig) {
    return function () {
        var state = getState();
        return proxyStateForModel(state, modelConfig);
    };
};

var handleWithContext = exports.handleWithContext = function handleWithContext(modelContext, handler) {
    var modelConfig = getModelConfig(modelContext);
    return function (_ref) {
        for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
            args[_key - 1] = arguments[_key];
        }

        var dispatch = _ref.dispatch,
            getState = _ref.getState;

        var storeApi = {};
        dispatch && (storeApi.dispatch = wrapDispatchWithModelContext(dispatch, modelConfig));
        getState && (storeApi.getState = wrapGetStateWithModelContext(getState, modelConfig));
        return handler.call.apply(handler, [null, storeApi].concat(args));
    };
};

var extendModel = exports.extendModel = function extendModel(absModel, options, mainModel) {
    if (!mainModel && options.namespace) {
        mainModel = options;
        options = {};
    }
    var _options = options,
        _options$prefix = _options.prefix,
        prefix = _options$prefix === undefined ? '' : _options$prefix;

    // absModel namespace 检查

    (0, _invariant2.default)(/^\$/.test(absModel.namespace), '[connectWithContext] Abstract model\'s namespace must start with `$`.');
    // 重复字段检查
    ['state', 'reducers'].forEach(function (key) {
        Object.keys(mainModel[key]).forEach(function (item) {
            var absModelProp = absModel[key][item];
            (0, _invariant2.default)(!absModel[key].hasOwnProperty(item)
            // 允许mainModel中把abstract model的state作为initial state整合，但仅支持引用类型的state以引用方式整合
            || absModelProp && (typeof absModelProp === 'undefined' ? 'undefined' : _typeof(absModelProp)) === 'object' && absModelProp === mainModel[key][item], '[connectWithContext] Main model `' + mainModel.namespace + '` has already defined `' + key + '.' + item + '`');
        });
    });
    return {
        namespace: mainModel.namespace,
        state: _extends({}, mainModel.state, !prefix ? absModel.state : Object.keys(absModel.state).reduce(function (ret, key) {
            ret[prefix + key] = absModel.state[key];
            return ret;
        }, {})),
        reducers: _extends({}, mainModel.reducers, !prefix ? absModel.reducers : Object.keys(absModel.reducers).reduce(function (ret, key) {
            ret[prefix + key] = function (state, action) {
                var prefixedState = Object.keys(absModel.state).reduce(function (ret, key) {
                    ret[key.substr(prefix.length)] = state[key];
                    delete ret[key];
                    return ret;
                }, _extends({}, state));
                var nextState = absModel.reducers[key](prefixedState, action);
                return Object.keys(absModel.state).reduce(function (ret, key) {
                    ret[prefix + key] = nextState[key];
                    delete nextState[key];
                    return ret;
                }, nextState);
            };
            return ret;
        }, {}))
    };
};

exports.default = function (app) {
    return function (getUIState) {
        for (var _len2 = arguments.length, connectArgs = Array(_len2 > 2 ? _len2 - 2 : 0), _key2 = 2; _key2 < _len2; _key2++) {
            connectArgs[_key2 - 2] = arguments[_key2];
        }

        var callbacks = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

        return function (UI) {
            return function (modelContext) {
                var ConnectedWithContext = function (_React$Component) {
                    _inherits(ConnectedWithContext, _React$Component);

                    function ConnectedWithContext() {
                        _classCallCheck(this, ConnectedWithContext);

                        return _possibleConstructorReturn(this, (ConnectedWithContext.__proto__ || Object.getPrototypeOf(ConnectedWithContext)).apply(this, arguments));
                    }

                    _createClass(ConnectedWithContext, [{
                        key: 'getChildContext',
                        value: function getChildContext() {
                            return modelContext ? {
                                model: _extends({}, this.context.model || {}, modelContext)
                            } : undefined;
                        }
                    }, {
                        key: 'render',
                        value: function render() {
                            var modelCtx = modelContext || this.context.model;
                            (0, _invariant2.default)(modelCtx, '[connectWithContext] No model context found!');

                            var Component = void 0;
                            var cachePath = [getUIState, callbacks, UI, JSON.stringify(modelCtx)];
                            Component = getCache(cachePath);
                            if (!Component) {
                                var modelConfig = getModelConfig(modelCtx);
                                var wrappedModelGetUIState = function wrappedModelGetUIState() {
                                    for (var _len3 = arguments.length, args = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
                                        args[_key3] = arguments[_key3];
                                    }

                                    var state = args[0],
                                        extArgs = args.slice(1);

                                    return getUIState.apply(undefined, [proxyStateForModel(state, modelConfig)].concat(_toConsumableArray(extArgs)));
                                };
                                var wrappedModelCallbacks = Object.keys(callbacks).reduce(function (ret, key) {
                                    ret[key] = function (_ref2) {
                                        var _callbacks$key;

                                        for (var _len4 = arguments.length, args = Array(_len4 > 1 ? _len4 - 1 : 0), _key4 = 1; _key4 < _len4; _key4++) {
                                            args[_key4 - 1] = arguments[_key4];
                                        }

                                        var dispatch = _ref2.dispatch,
                                            getState = _ref2.getState;

                                        (_callbacks$key = callbacks[key]).call.apply(_callbacks$key, [null, {
                                            dispatch: wrapDispatchWithModelContext(dispatch, modelConfig),
                                            getState: wrapGetStateWithModelContext(getState, modelConfig)
                                        }].concat(args));
                                    };
                                    return ret;
                                }, {});

                                Component = app.connect.apply(app, [wrappedModelGetUIState, wrappedModelCallbacks].concat(connectArgs))(UI);
                                setCache(cachePath, Component);
                            }

                            return _react2.default.createElement(Component, this.props);
                        }
                    }]);

                    return ConnectedWithContext;
                }(_react2.default.Component);

                ConnectedWithContext.contextTypes = {
                    model: _react2.default.PropTypes.any.isRequired
                };

                if (modelContext) {
                    ConnectedWithContext.childContextTypes = {
                        model: _react2.default.PropTypes.any.isRequired
                    };
                }

                return ConnectedWithContext;
            };
        };
    };
};